// header file for C++ automaton representation
//

// note: all variables in hybrid system are global
// there are num_vars of them


// information will be passed via pointers
// initially form the structure as linked lists (<class>_list)
// then regroup into arrays with pointer to head for faster access
// <class> * <class>_head
//
// shall transitions be stored contiguously per location or shared? 
//    contiguous is faster access.
//    shared is less memory

// There are global lists for slope_vectors, the global var the_slopes
//    and for the matrices, matrix_hd within the automaton class
//
// All references to these vectors and matrices are via ptrs to the global
//    elements on the list.  This is because many vectors and matrices
//    are used more than once, and should be stored only once.
//    If the check for duplication is not done now, that's OK, but please
//    make sure that it is easy to add it later, otherwise it defeats
//    the purpose of having a global list, and using ptr reference.
// Transitions, and assignments are kept locally within the automaton, with
//    ptrs as necessary to the matrices for guards and reassignment of vars
//

// the routines in convert.C/H will be used to convert C internal structures
// into C++ internal structures, for the POST information only.
// Greg, you can ignore the PRE stuff.  I will do this.
#include "poly.H"            // include the polyhedral library
                             // beware name conflicts with parser structures
                           
#include "include.h"
#include "basic.h"
#include "omatrix.h"
#include "opoly.h"
#include "convert.H"

#ifndef __AUTOMATON_H
#define __AUTOMATON_H

// all code for using type polyhedron for library, now refers to opoly
typedef opoly polyhedron;

#define MAX_VARS 150
#define MAX_PROCS 40

// forward class declarations

class location;
class automaton;
class region_array;


// take this from the C code
//
extern int var_count;   // global number of variables

extern int debug_magic_nr; // global for debugging purposes

void global_free_str(char *str);

void global_free( void *p);

void trace_out(char *tracestring, int traceint);

// put in *mp the C++ matrix for m 
void
Cpp_matrix(matrix *mp, struct _p_matrix m);


char *
Bool_output(Bool b);


void
clear_ptr_int_array(int **a, int nr_elts);

coeff abs_val(coeff a);


coeff min_of(coeff a, coeff b);

coeff max_of(coeff a, coeff b);

coeff lcm_coeff(coeff x, coeff y);

// return index of posn of key in the int array arr.
// return -1 if not in the array
int 
pos_in_int_array(int key, int *arr, int size);

// list of names of the variables.
// start list from item 0.
extern char *varnames[];
// whether each var should be eliminated or not in final output
// info determined by elim_proc_names
extern Bool elim_var_index[MAX_VARS];
extern Bool elim_proc_index[MAX_PROCS];
extern char *infilename;

extern int print_mode;  // 0 = minimal output, 1 = show all reach iterations
extern int format_mode; // 0 = print across page, 1 = print down

/* was in basic but slowed down code tremendously */

/* flags */
extern Bool check_automaton_flag;
extern Bool reach_with_edges;
extern Bool did_reach_with_edges;
extern int max_reg_id;
extern Bool generate_forward_trace;
extern Bool reached_forward;
extern Bool generating_trace;



Bool test_ok_to_generate(automaton *a);

/*   ======== CLASSES ======== */



class assignment {
public:
  int nr_vars;               /* number of variables being reassigned */
  int * indices;             /* pointer to list of nr_var indices    */
                             /* numbering starts at 0, for variable ref too */
                             /* i.e. indices[1] = 4, means the second */
                             /* reassigned variable is 5th declared var*/
                             /* for inverse ass., store positive index for */
                             /* var used in old form and primed form */
                             /* (store such vars in extra columns */

                             /* use nr_new_vars, etc, when var is used in */
                             /* for primed form, but not  unprimed */
                             /*  form (store such vars in regular columns) */
  int nr_new_vars;
  int *new_indices;
  matrix * ass_constraints;  /* pointer to matrix for the assignment */
  

  assignment(int nbvars, int *ind, matrix * mp);
  assignment();

  void operator = (const assignment& ass);
  assignment make_inverse_ass(); /* ass_constr is ptr to some stored matrix */
  void set_matrix_p(matrix *);   /* set ass_constr to matrix in list */
                                 /* free previous ass constr matrix pted to */

  void print();
  void free();   /* free ptrs, and also data for indices, but NOT for matrix */
};


class transition {
public:
  location *src;             /* src, and dest viewed in forward direction */
  location *dest;
  char *sync_label;
  matrix * enabling_constr;  /* always the guard in forward direction */
  assignment ass;
  Bool is_asap_trans;        // = "is ASAP transition"
                             // meaning taken as soon as enabled.
                             // cannot have any guard condition or assignment
                             // constraints (?)
                             // no cts evolution taken out of the source locn

  transition();
  transition(location *src, location *dest, char *sync_labl,
	     matrix *gd, Bool is_asap, assignment *assgt);

  matrix * fw_enabling_matrix();   /* enabling matrix */
//   void operator = ( transition tr);

 // added from last freeze
  void free();                // free ptrs, not data
  void print();
  void brief_print();        // don't print assignment info
  void label_print();        // print just the sync label for trace generation
};

/* =================== Lists used in creating structures ================ */

struct slink {
  slink *next;
  slink() {next = 0;}
  slink(slink * p) {next = p;}
};

class slist_base {
  slink* last;         // last->next is head of list
public:
  void insert(slink *);   // add to head of list
  slink * get();         // remove and return head of list

  void clear() { last = 0; }
  slist_base() { last = 0; }
  slist_base(slink* a) { last = a->next = a; }

  friend class slist_base_iter;
};


template<class T>
class Islist : private slist_base {
public:
  void insert(T* a) { slist_base::insert(a); }
  T* get()          { return (T*) slist_base::get(); }
};


template<class T>
struct Tlink : public slink {
  T info;
  Tlink (const T& a): info(a) {}
};

template<class T>
class Slist : private slist_base {
public: 
  void insert(const T& a)
    {slist_base::insert(new Tlink<T>(a)); }
  T get();
};

template<class T>
T Slist<T>::get()
{
  Tlink<T>* lnk = (Tlink<T>*) slist_base::get();
  T i = lnk->info;
  delete lnk;
  return i;
};

class slist_base_iter {
  slink* ce;      // current element
  slist_base* cs; // current list
public:
  inline slist_base_iter(slist_base& s);
  inline slink* operator()();
};


/* ============== automaton lists =========================== */

//
// Warning: usage of lists.
// This is nasty. Insertion assumes there is a head item already.
// Thus to initialize a list, assign it the first item.
// Then continue doing insertions.
// NO list deletion yet!
//

class matrix_info {
public:
  matrix m;
  matrix_info * next;

  matrix_info * insert(matrix_info *mip); 
                                     //insert matrix pted to as head of list
                                     // also return head
  matrix_info(const matrix & m);
  matrix_info(int nr_cols); /* return an item with TOP for nr columns */
  void print();
};


// my use lists of transitions in forming pre, etc.
// though new version looks ahead to count number needed and forms array.
class transition_info  {
 public:
  transition trans;
  transition_info *next;

  transition_info * insert(transition_info *tip);
  transition_info();
  transition_info(transition& tr);
};


// why do i need lists of assignments?
class assignment_info  {
 public:
  assignment *asst;
  assignment_info *next;

  assignment_info * insert(assignment_info *aip);
  assignment_info(assignment *ass);
};



/* =================== Locations ================ */


typedef vector slope_vector;

class slope_vector_info {
 public:
   vector v;
   slope_vector_info *next;

   slope_vector_info();
   slope_vector_info(slope_vector& vec);
   slope_vector_info * insert(slope_vector_info *svip);
   void print();
 };

typedef slope_vector_info *slope_vectors_info;


class location {
 public:
  char *name;
  char *copy_of_name;  

  int nr_slopes;
  int           locn_id;
  polyhedron    *post_slopes_matrix;  // ptr to poly for post slopes
  polyhedron    *pre_slopes_matrix;  // ptr to poly for pre slopes
  slope_vector  **post_slopes;      // ptr to array of ptrs to slopes
  slope_vector  **pre_slopes;       // since many slopes will be shared
  int           nr_post_trans;
  int           nr_pre_trans;
                                    // store as ptrs to array of transitions
  transition *  post_trans;         
  transition *  pre_trans;          
  transition_info * pre_trans_list; // not need, since count # pre-trans first
                                    // so can allocate array directly
  Bool          asap;               // = "is ASAP location"
                                    // means that there is an ASAP transition
                                    // out of this location.
                                    // This implies do NOT take cts successors
                                    
  matrix *      invariant;
  automaton*    autp;

  location();
  location(char * l_name);
  location(const location & locn);
  void          set_post_vectors();  // ????
  void          set_pre_vectors();
  // assuming the transitions are set, fix asap info for the locations
  void          set_asap_from_trans();
  slope_vector  **get_slope_vectors_for(enum directions d);
  void          restrict_to_invariant();  // ???
  void          print();
  void          name_print();
  void          free();

  // copy name to copy_of_name
  void copy_loc_name();
  // copy copy_of_name to name
  void restore_loc_name();
};

/* ==================== component info =================== */

class component_info {
public:
  char *automaton_name;
  int nr_locations;
  char **locn_names;   // stored in order of locn_ids

  component_info();
  component_info(automaton *autp);  // constructs object with info for this aut
  void print();
  void free();
};


/*  =========================== automaton ====================== */

class automaton {
 public:
   char *name;
   int        nr_slopes;
   slope_vector *slopes;         // ptr to array of slopes in free space
   int        nr_locations;
   location * locations; 
   matrix      * matrix_store;    // store of actual matrices in free space
   int         nr_matrices_in_store;
                    // array of length nr_locations.  
                    // entry true iff  matrix in matrix_store has "original" 
                    // copy of the data 
   Bool        *is_orig_matrix;

   matrix_info * matrix_hd;
   int         nr_sync_labels;
   char        **sync_label_names;   /* ptr to array of sync label names */

   int         *id_slope_index;      // id's (from C) for vectors
   int         *id_matrix_index;    // id's (from C) for matrices
                                     // i.e. ids_for_matrices[3] = 10 means
                                     // the matrix with id 3 is in 10th posn
                                     // in matrix_store
   location    *init_location_p;       // ptr to initial location
   matrix      * init_region_p;        // ptr to constraint matrix for 
                                       // initial region
   int         slope_lcm;            // lcm for all slope vectors
   Bool        pre_is_valid;         // whether info for pre- is correctly
                                     // stored.


                                      // how many component aut used to make
   int nr_components;                 // this automaton
                                      // info re each component (in order)
   component_info *comp_inf_p;        // store as ptr to array

   automaton();
   automaton(int s);             // create an automaton with s locations
   automaton(char *nom, int s);  // create an automaton nom with s locations
   automaton *copy_of();  // copy automaton, ptrs only, not data
   automaton(struct automata_list C_automaton); // create from a C_automaton
   
   // used for converting from C to C++ structure
   void copy_sync_labels(struct  automata_list C_a);

   void copy_slopes(struct automata_list C_automaton); // create space for and
                                                  // copy global slope list
                                                  // and ids_for_slopes

   void copy_matrices(struct automata_list C_automaton); 
                                          // create space for and
                                          // copy global matrix list
                                          // and ids_for_matrices

   void copy_locations(struct automata_list C_automaton);
                              // copy data into the locations.
                              // this includes copying the transitions out of 
                              // each location.  and creating new matrices for
                              // assignments. and setting ptrs, where C 
                              // structure had ID numbers.

   void copy_init_info(struct automata_list C_automaton);
                              // copy over info for the init_region and locn
                              // assumes lists of matrices, locn copied already

   void adjust_locn_ids();    // set all locn_id's to be the index
                              // of the location.

   void free_id_vectors();    // free space for ids_for_slopes/matrices

                              // given trp and C_automaton, 
                              // set up last 3 parameters
   void set_up_assignment(struct info_transition *trp, 
			  struct automata_list C_automaton,
			  int *nbvars, int **indp, matrix **ass_mp);

   void new_set_up_assignment(struct info_transition *trp, 
			  struct automata_list C_automaton,
			  int *nbvars, int **indp, matrix **ass_mp);
   void old_set_up_assignment(struct info_transition *trp, 
			  struct automata_list C_automaton,
			  int *nbvars, int **indp, matrix **ass_mp);


                              // return 1 if ass is single equality
                              // returns 2 otherwise (lower and upper bound)
                              // currently no way for single lower/upper bound?
   int nr_constraints_in(struct info_assign *asp, struct automata_list C_a);

   location *dest_location_ptr(char *dest_name);
                              // set all the ASAP info for locns
                              //   based on transitions' asap info
   void set_location_asap();

   // for checking 'syntax'/structure of automaton
   int check_automaton();

   // for forming products
   automaton compose(automaton aut);    // compose this object with aut

   // for forming products, using control reachability
   automaton reach_compose(automaton aut);    // compose this object with aut

   // copy data, not ptrs to data, from *aut1 to object
   void copy_comp_info(automaton *aut1);

   // add a single component's locn data to object, put in posn for comp_nr
   void add_comp_info(int comp_nr, automaton *autp);

   // make the component info correct (used in forming product with *autp)
   void set_up_comp_inf(automaton *aut1, automaton *aut2);

   // given ptrs to two automaton, set up nr_sync_labels, and the actual labels
   void set_up_alphabet(automaton *aut1, automaton *aut2);

   // set up info for locations, except for the transitions
   void set_up_locns(automaton *aut1, automaton *aut2);


   // set up info for locations, using reachability info
   void set_up_reach_locns(automaton *aut1, automaton *aut2,
				    int nr_reached, int *is_reached);

   void set_product_slopes(int i,int j, automaton *aut2, 
			 coeff **slope_1_min, coeff **slope_1_max,
			   int **undecl_1,
			 coeff **slope_2_min, coeff **slope_2_max,
			   int **undecl_2,
			 int **vecindex);

   void set_up_product_invariant(int i, int j, 
				 automaton *aut1, automaton *aut2);
   void set_up_reach_product_invariant(int i, int j, 
				 automaton *aut1, automaton *aut2,
				       int index);
   
   // set up transitions, except use indexptrs to the matrices (ref to 
   //  matrix_store index posn).
   void set_up_trans(int i, int j, automaton *aut1, automaton *aut2,
		     int **actualguardindexp, int **actualassindexp);

   void set_up_reach_trans(int i, int j, automaton *aut1, automaton *aut2,
		     int **actualguardindexp, int **actualassindexp,
			   int index, int *is_reached);


   // put in correct ptrs to matrices, based on the index info
   void adjust_ptrs_to_matrices(int i, int j,  automaton *aut2,
		     int **actualguardindex, int **actualassindex);
   void adjust_ptrs_to_matrices_reach(int i, int j,  automaton *aut2,
		     int **actualguardindex, int **actualassindex,
				      int index);


   // assumes the locations of product already there.
   void set_up_init_info(automaton *aut1, automaton *aut2);
   void set_up_init_info_reach(automaton *aut1, automaton *aut2, 
			       int *is_reached);
   

   // put this vector in global list of vectors, add it to slopes for
   // locations [index]
   void install_vector(slope_vector sv, int index, int **vecindex);

   // put new matrix in the matrix_store
   // size contains size of matindex[index] array
   // 5/9/95: changed to take addr of ptr to matrix.  If matrix known to appear
   // already then free matrix, and set ptr to previous occurrence of matrix
   // in the store.
   // space_flag indicates how prudent to be about saving space
   // space_flag = 0 => install matrix regardless
   //            = 1 => search beyond invariant matrices
   //            = 2 => search over all matrices
   void install_matrix(matrix **mpaddr, int index, int **matindex, int size,
		       int space_flag);
 
   // add a post_trans, number tnr of aut1's location id1
   // allocate space for it in post_trans
   // store guard and assignment matrices in matrix_store, and update
   // info in the matindices
   // whichaut =1 or 2 for source  of unsynch (single) trans
  void add_trans(int whichaut,
		 int tnr, automaton *aut, int locn_nr_1, int locn_nr_2,
		 int size2, int index,
		  int **actualguardindex, int **actualassindex);
  void add_reach_trans(int whichaut,
		       int tnr, automaton *aut1, automaton *aut2, 
		       int locn_nr_1, int locn_nr_2,
		       int size2, int index,
		       int **actualguardindex, int **actualassindex,
		       int *is_reached);
     


   // add a synchronized transition.  using
   // trans tnr1 of aut1's id1 locn, and trans tnr2 of aut2's id2 locn
   // store indexptrs in actualguardindex and actualassindex
   void add_sync_trans(int tnr1, int tnr2,automaton *aut1, automaton *aut2,
	       int id1, int id2, 
	       int **actualguardindex,
	       int **actualassindex) ;
   void add_sync_reach_trans(int tnr1, int tnr2,
			     automaton *aut1, automaton *aut2,
			     int id1, int id2, 
			     int **actualguardindex,
			     int **actualassindex,
			     int index,
			     int * is_reached) ;


   // use the "reach" routines in region.H
   // verification
   void simple_reachability(int dir);  // perform simple reachability analysis
                                //  in requested direction
                                // and print out reachable regions.

   //  use the "reach" routines in region.H
   // this version is much cleaner and uses region arrays
   // still no plugs of memory leaks though.
// how to get forward ptr?
   region_array *new_reach(enum directions dirn);


   // set up elim loc by list and elim var by list info.
   // the parser already sets up the elim_all_locns, elim_all_vars vars
   // must convert from names in elim_var_names_list and elim_proc_names_list
   // to indices, and Bool arrays.
   void set_elim_info();


   // copy locn name  to copy_of_name, for use with restore_location_names
   void copy_location_names();
   // restore the names of the locations, using the component info.
   // needed for after the routiens that change the names of the locations
   // for simplification.  nasty.
   void restore_location_names();

   // general useful utilities
   void           print();  
   void           print_size();      // output #locns, #pretrans, #posttrans
                                     // # matrices in store, #vectors
   void           set_matrix_head(); // initialize matrix to have TOP only
   void           set_pre_vectors(); // obsolete, see set_pre_slope_matrices
   void           set_pre_slope_matrices(); // later make this protected
   void           set_pre_trans(); // later make this protected
   void           set_pre();      // assumes post- info already there
   void           add_inverse(transition & tr); // add inverse transition.
                                       // space in pre_trans already allocated
   matrix_info *store_matrix_from_assignment(assignment inv_ass);


   // freeing memory
   // this one used in compose.C
   void           free_actual_indexes(int **actualguardindex,
				      int **actualassindex);

   // freeing mem for actual automaton
   void           free_component_info();
   void           free_matrix_store();
   void           free_slopes();
   void           free_locations();
   void           free(); // free matrix_store and vectors, and all transitions


 }; // end automaton class declaration



class automaton_info {
public:
   automaton *a;    // ptr to the automaton in this link
   automaton_info *next;

   automaton_info();
   automaton_info(automaton aut);      // copy constructor
                                       // copy ptrs not data!
   automaton_info(struct automata_list C_automaton); 

   automaton_info *insert(automaton_info *aut_inf);

                                       // composes the entire list of automata
                                       // returns singleton list w. aut prod.
   automaton_info *compose();
   automaton_info *reach_compose();    // generates control reachable locs only
   void print();
 };

/* ====================  global vars ============================== */

extern slope_vectors_info the_slopes;  // global list of all slopes
extern automaton global_a;



#endif




