/* routines related to specifications as they're parsed in (all in C) */

/*
 last change:
  8/14/00. 
  in store_expr_in_empty_Boolexpression() correctly incremented
      ptr into Bool expression stack before assignment of expression.
      i.e., added the "Boolexpr_depth++" before 
      "Boolexpr_stack[Boolexpr_depth] = Boolexpr;"
*/


#include "spec.h"

/* global region table */
struct region_name_table reg_name_table;


/* a outer level statement is stored in a list with head global_stmt */
/*                                                                   */
/* routines use the global variables parse_stmt and parse_expr       */
/* for temporary values of stmts  or expressions currently being parsed.  */
/* need a stack for statements, and expressions, since parsed recursively */
/* also need to record where in the stack the current list of statements  */
/* starts, so that when the list is done with, they can be pasted         */
/* together as a list of statements                                       */

struct statement *global_stmt = (struct statement *) NULL;

struct expression * expr_stack[MAX_EXPR_DEPTH];
struct statement *  stmt_stack[MAX_STMT_STACK_SIZE];
struct Boolexpression * Boolexpr_stack[MAX_BOOLEXPR_DEPTH];
int stmt_depth_stack          [MAX_STMT_DEPTH];
int stmt_stack_nr = -1;
int expr_depth = -1;
int stmt_depth = -1;
int Boolexpr_depth = -1;

int old_spec_style = 0;   /* assume using new spec style */

/* ======================================== */

void 
init_region_name_table(){
  int i;
  reg_name_table.size = 0;
  for (i=0;i<MAX_REGION_NAMES;i++) {
    reg_name_table.names[i] = (char *) NULL;
  }
}

void 
store_region_name(char *rname){
  if (reg_name_table.size == MAX_REGION_NAMES) {
    fprintf(stderr,"Too many region names declared: max = %d\n",
	    MAX_REGION_NAMES);
    exit(1);
  }
  reg_name_table.names[reg_name_table.size] = str_preserve(rname);
  reg_name_table.size++;
}

void
check_region_name(char *rname) {
  if (string_pos(rname,reg_name_table.names,reg_name_table.size) == -1) {
    fprintf(stderr,"Variable %s not declared. aborting\n",rname);
    exit(1);
  }
}



/* ======================================== */

/* constructor for expressions */
struct expression *
get_new_expression() {
  struct expression * expr = 
    (struct expression *) malloc (sizeof(struct expression));
  expr->type = PREDICATE;
  expr->rname = (char *) NULL;
  expr->rindex = -1;
  expr->expr1 = (struct expression *) NULL;
  expr->expr2 = (struct expression *) NULL;
  expr->stmt = (struct statement *) NULL;
  expr->pred = (struct convex_predicate *) NULL;
  return (expr);
}


struct statement *
get_new_statement() {
  struct statement * stmt =
    (struct statement *) malloc (sizeof(struct statement));
  stmt->type = ASSIGN_STMT; /* randomly chosen default type */
  stmt->rname = (char *) NULL;
  stmt->rindex = -1;
  stmt->expr =  (struct expression *) NULL;
  stmt->elim =  (struct elim_info *) NULL;
  stmt->thenstmt =  (struct statement *) NULL;
  stmt->elsestmt =  (struct statement *) NULL;
  stmt->next = (struct statement *) NULL;
  return(stmt);  
}

struct Boolexpression *
get_new_Boolexpression() {
  struct Boolexpression * Boolexpr = 
    (struct Boolexpression *) malloc (sizeof(struct Boolexpression));
  /* set the default to true */
  Boolexpr->type = BE_INEQ;
  Boolexpr->left_expr = (struct expression *) NULL;
  Boolexpr->right_expr = (struct expression *) NULL;
  Boolexpr->Boolexpr1 = (struct Boolexpression *) NULL;
  Boolexpr->Boolexpr2 = (struct Boolexpression *) NULL;
  Boolexpr->optype = EQ;
  return( Boolexpr );
}


struct elim_info *
get_new_elim_info() {
  struct elim_info *elim =
    (struct elim_info *) malloc (sizeof(struct elim_info));
  /* set up defaults */
  elim->elim_all = 0;
  elim->name_list = (struct var_name *) NULL;
  return(elim);
}

/* ======== printing routines ============= */

void
print_expression_type(enum expr_type exp_type) {
  switch (exp_type) {
  case REGION_NAME: printf("NAME"); 
    break;
  case AND: printf("AND");
    break;
  case OR: printf("OR");
    break;
  case WEAKDIFF: printf("WEAKDIFF");
    break;
  case DIFF: printf("DIFF");
    break;
  case PRE_EXPR: printf("PRE"); 
    break;
  case POST_EXPR: printf("POST"); 
    break;

  case OMIT_EXPR: printf("OMIT"); 
    break;
  case PREDICATE: printf("PRED:"); 
    break;
  case REACHFORWARD: printf("REACHFORWARD:"); 
    break;
  case REACHBACKWARD:  printf("REACHBACKWARD:");
    break;
  case RECURSION:  printf("RECURSION:"); 
    break;
  case HIDE_EXPR:  printf("HIDE:"); 
    break;
  case HULL_EXPR:  printf("HULL:"); 
    break;
  default:
     fprintf(stderr,"Illegal call of print expr, illegal expr type\n");
    ansi_fflush(stderr);
    exit(1);
  }
  ansi_fflush(stdout);
}





void
print_elim_info(struct elim_info *elim) {
  switch (elim->elim_all)  {
  case 2:
    printf(" ALL\n");
    break;
  case 1:
    printf("NON PARAMETERS\n");
    break;
  case 0: default:
    print_vars(elim->name_list);
    break;
  }
  ansi_fflush(stdout);
}

void
print_optype(int optype) {
  switch (optype) {
  case LE:
    printf(" <= \n");
    break;
  case GE:
    printf(" >= \n");
    break;
  case EQ:
    printf(" = \n");
    break;
  case GT:
    printf(" > \n");
    break;
  case LT:
    printf(" < \n");
    break;
  case NEQ:
    printf(" != \n");
    break;
  default:
    fprintf(stderr,"Illegal call of print_optype with argument %d\n",optype);
    ansi_fflush(stderr);
  }
}

void
print_weakoptype(enum Boolexpr_type type) {
  switch(type) {
  case BE_WEAKLE:
    printf("WEAKLE\n");
    break;
  case BE_WEAKEQ:
    printf("WEAKEQ\n");
    break;
  case BE_WEAKGE:
    printf("WEAKGE\n");
    break;
  default:
    fprintf(stderr, "Illegal type %d in call of print_weakoptype\n", type);
    exit(1);
  }
}

void
print_Boolexpression(struct Boolexpression * Boolexpr) {
  switch (Boolexpr->type) {
  case BE_AND:
    printf("(");
    print_Boolexpression(Boolexpr->Boolexpr1);
    printf("Bool-AND\n");
    print_Boolexpression(Boolexpr->Boolexpr2);
    printf(")\n");
    break;
  case BE_OR:
    printf("(");
    print_Boolexpression(Boolexpr->Boolexpr1);
    printf("Bool-OR\n");
    print_Boolexpression(Boolexpr->Boolexpr2);
    printf(")\n");
    break;
  case BE_NOT:
    printf("NOT (");
    print_Boolexpression(Boolexpr->Boolexpr1);
    printf(")\n");
    break;
  case BE_EMPTY:
    printf("IS EMPTY ?? (");
    print_expression(Boolexpr->left_expr);
    printf(")\n");
    break;
  case BE_INEQ:
    print_expression(Boolexpr->left_expr);
    print_optype(Boolexpr->optype);
    print_expression(Boolexpr->right_expr);
    break;
  case BE_WEAKLE: case BE_WEAKEQ: case BE_WEAKGE:
    print_expression(Boolexpr->left_expr);
    print_weakoptype(Boolexpr->type);
    print_expression(Boolexpr->right_expr);
    break;
  default:
    fprintf(stderr,"Illegal call,  printing Bool expr\n");
    ansi_fflush(stderr);
  }
  ansi_fflush(stdout);
}


/* print a single statement */
void
print_statement(struct statement *stmt) {
 if (stmt == NULL) {
   printf("Cannot print a NULL statement\n");
   ansi_fflush(stdout);
   exit(1);
 }

 switch( stmt->type) {
 case FREE_STMT:
   printf("FREE %s", stmt->rname);
   break;
 case PRINT_STMT:
   printf("PRINT ");
   print_expression(stmt->expr);
   break;
 case PRINT_SIMPLIFIED_STMT:
   printf("PRINTSIMPLIFIED ");
   print_expression(stmt->expr);
   printf(" HIDING " );
   print_elim_info(stmt->elim);
   break;
 case PRINT_STR_STMT:
   printf("PRINTSTRING %s", stmt->rname);
   break;
 case PRINTSIZE_STMT:
   printf("PRINTSIZE %s", stmt->rname);
   break;
 case TRACE_STMT:
   printf("TRACE TO ");
   print_expression(stmt->expr);
   break;
 case ASSIGN_STMT:
   printf("ASSIGN %s TO ", stmt->rname);
   print_expression(stmt->expr);
   break;
 case IF_THEN_STMT:
   printf("IF ");
   print_Boolexpression(stmt->Boolexpr);
   printf("THEN\n");
   print_statements(stmt->thenstmt);
   break;
 case IF_THEN_ELSE_STMT:
   printf("IF ");
   print_Boolexpression(stmt->Boolexpr);
   printf("THEN\n");
   print_statements(stmt->thenstmt);
   printf("ELSE\n");
   print_statements(stmt->elsestmt);
   break;
 case WHILE_STMT:
   printf("WHILE ");
   print_Boolexpression(stmt->Boolexpr);
   printf("DO\n");
   print_statements(stmt->loopstmt);
   printf("ENDWHILE\n");
   break;
 }
 printf("\n");
 ansi_fflush(stdout);
 
}


/* print a list of statements */
void
print_statements(struct statement *stmt) {
  if (stmt != NULL) {
    print_statement(stmt);
    print_statements(stmt->next);
  }
}

		    

void
print_expression(struct expression *expr) {
  if (expr == NULL) {
    return;
  }

  print_expression_type(expr->type); printf(" ");
  switch (expr->type) {
  case REGION_NAME:
    printf(" %s\n", expr->rname);
    break;
  case AND: case OR: case WEAKDIFF: case DIFF:
    print_expression(expr->expr1);
    print_expression(expr->expr2);
    break;
  case PRE_EXPR: case POST_EXPR:
    print_expression(expr->expr1);
    break;
  case PREDICATE: 
    print_convex_predicate(expr->pred);
    break;
  case REACHFORWARD: case REACHBACKWARD:
    printf(" FROM ");
    print_expression(expr->expr1);
    break;
  case RECURSION:
    printf(" OVER ");
    printf(" %s FROM ", expr->rname);
    print_expression(expr->expr1);
    printf(" USING " );
    print_statements(expr->stmt);
    break;
  case HIDE_EXPR:
    print_elim_info(expr->elim);
    printf(" IN ");
    print_expression(expr->expr1);
    break;
  case HULL_EXPR:
    print_expression(expr->expr1);
    break;
  default:
    fprintf(stderr,"Illegal call of print expr, illegal expr type\n");
    ansi_fflush(stderr);
    exit(1);
  }
  ansi_fflush(stdout);
}

/* ================================================== */
/*                handle statements                   */
/* ================================================== */


void
init_statements() {
  stmt_depth++;
  if (stmt_depth>=MAX_STMT_DEPTH) {
    printf("Aborting: max. nested statement depth of %d has been exceeded\n", 
	   MAX_STMT_DEPTH);
    exit(1);
  }
  /* inc one past stmt_stack_nr, i.e. where the first stmt *will* be */
  stmt_depth_stack[stmt_depth] = stmt_stack_nr+1;
}


/* link together all statements from stmt_depth_stack[stmt_depth] to */
/* stmt_stack_nr, leave resulting statement in first position */
/* update stmt_stack_nr, but not stmt_depth since the parser uses this   */
/* info to get the statement list */
void
store_statements() {
  int i;
  struct statement *stmt = stmt_stack[stmt_depth_stack[stmt_depth]];

/*
fprintf(stderr,"\n\n\nin store statements\n\n");
for (i=0;i<=stmt_stack_nr;i++) {
  printf(">>> %d <<< ",i);
  print_statement(stmt_stack[i]);
}
ansi_fflush(stdout);
ansi_fflush(stderr);
fprintf(stderr,"\n\n stmt stack nr = %d, stmtdepth = %d, stmtdep[stdep] = %d\n\n", stmt_stack_nr, stmt_depth, stmt_depth_stack[stmt_depth]);
ansi_fflush(stdout);
ansi_fflush(stderr);
*/


  for (i=stmt_depth_stack[stmt_depth]+1;i<=stmt_stack_nr;i++) {
    stmt->next = stmt_stack[i];
    stmt_stack[i] = (struct statement *) NULL;
    stmt = stmt->next;
  }
  stmt_stack_nr = stmt_depth_stack[stmt_depth];

/*
fprintf(stderr,"\n\n\nin store statements\n\n");
for (i=0;i<=stmt_stack_nr;i++) {
  printf(">>> %d <<< ",i);
  print_statements(stmt_stack[i]);
}
ansi_fflush(stdout);
ansi_fflush(stderr);
fprintf(stderr,"\n\n stmt stack nr = %d, stmtdepth = %d, stmtdep[stdep] = %d\n\n", stmt_stack_nr, stmt_depth, stmt_depth_stack[stmt_depth]);
ansi_fflush(stdout);
ansi_fflush(stderr);
*/


}


void
init_statement(enum stmt_type type) {
  stmt_stack_nr++;
  if (stmt_stack_nr >= MAX_STMT_STACK_SIZE) {
    printf("Aborting: number statements exceeds predefined limit of %d \n",
      MAX_STMT_STACK_SIZE);
    exit(1);
  }
  stmt_stack[stmt_stack_nr] = get_new_statement();
  stmt_stack[stmt_stack_nr]->type = type;
}

/*      ??    */
void
store_statement() {
  
}


/* specifics for individual types of statement */

void
store_name_in_stmt(char *name) {
  check_region_name(name);
  stmt_stack[stmt_stack_nr]->rname = str_preserve(name);
  stmt_stack[stmt_stack_nr]->rindex = 
     string_pos(name,reg_name_table.names,reg_name_table.size);
}

void
store_outputstr_in_stmt(char *outputstr) {
  stmt_stack[stmt_stack_nr]->rname = str_preserve(outputstr);
}  

void
store_expr_in_stmt() {
   stmt_stack[stmt_stack_nr]->expr = expr_stack[expr_depth];
/*
   print_expression(expr_stack[expr_depth]);
*/
   expr_stack[expr_depth] = NULL;
   expr_depth--;
 }

void
init_elim_var_info() {
  elim_all_vars = 0;
  had_elim_v_spec = 0;
  elim_vars_by_list = 0;
  elim_var_list = (struct var_name *) NULL;
}

void
init_elim_loc_info() {
  elim_all_locns = 0;
  had_elim_l_spec = 0;
  elim_locs_by_list = 0;
  elim_proc_list = (struct var_name *) NULL;
}

void
store_elim_loc_info() {
  stmt_stack[stmt_stack_nr]->elim = get_new_elim_info();
  stmt_stack[stmt_stack_nr]->elim->elim_all = elim_all_locns;
  stmt_stack[stmt_stack_nr]->elim->name_list = elim_proc_list;
  elim_proc_list = NULL;
}


/* ================================================== */
/*                handle expressions                  */
/* ================================================== */


void
init_expression(enum expr_type type) {
  expr_depth++;
  if (expr_depth >= MAX_EXPR_DEPTH) {
    printf("Aborting: expression depth exceeds limit of %d\n", 
	   MAX_EXPR_DEPTH);
    exit(1);
  }
  expr_stack[expr_depth] = get_new_expression();
  expr_stack[expr_depth]->type = type;
}

/*  ??  */
/* let parser do the work depending on how the expressions is to be used */
void
store_expression() {

}


void
store_elim_var_info() {
  expr_stack[expr_depth]->elim = get_new_elim_info();
  expr_stack[expr_depth]->elim->elim_all = elim_all_vars;
  expr_stack[expr_depth]->elim->name_list = elim_var_list;
  elim_var_list = NULL;
}

/* specifics for individual types of expressions */

void
store_name_expression(char *name) {
  check_region_name(name);
  expr_stack[expr_depth]->rname = str_preserve(name);
  expr_stack[expr_depth]->rindex = 
    string_pos(name,reg_name_table.names,reg_name_table.size);
  expr_stack[expr_depth]->type = REGION_NAME;
}


void
store_name_in_expression(char *name) {
  check_region_name(name);
  expr_stack[expr_depth]->rname = str_preserve(name);
  expr_stack[expr_depth]->rindex = 
    string_pos(name,reg_name_table.names,reg_name_table.size);
}


void
store_first_expression() {
  expr_stack[expr_depth-1]->expr1 =  expr_stack[expr_depth];
  expr_stack[expr_depth] = NULL;
  expr_depth--;
}

void
store_second_expression() {
  expr_stack[expr_depth]->expr1 =  expr_stack[expr_depth];
  expr_stack[expr_depth] = NULL;
  expr_depth--;
}

void
store_and_expression() {
  struct expression *expr = get_new_expression();
  expr->expr1 =  expr_stack[expr_depth-1];
  expr->expr2 = expr_stack[expr_depth];
  expr->type = AND;
  expr_depth--;
  expr_stack[expr_depth] = expr;
}

void
store_or_expression() {
  struct expression *expr = get_new_expression();
  expr->expr1 =  expr_stack[expr_depth-1];
  expr->expr2 = expr_stack[expr_depth];
  expr->type = OR;
  expr_depth--;
  expr_stack[expr_depth] = expr;
}

void store_weakdiff_expression() {
  struct expression *expr = get_new_expression();
  expr->expr1 =  expr_stack[expr_depth-1];
  expr->expr2 = expr_stack[expr_depth];
  expr->type = WEAKDIFF;
  expr_depth--;
  expr_stack[expr_depth] = expr;
}

void store_diff_expression() {
  struct expression *expr = get_new_expression();
  expr->expr1 =  expr_stack[expr_depth-1];
  expr->expr2 = expr_stack[expr_depth];
  expr->type = DIFF;
  expr_depth--;
  expr_stack[expr_depth] = expr;
}
  
void store_complement_expression() {
  struct expression *expr = get_new_expression();
  expr->expr1 =  expr_stack[expr_depth-1];
  expr->type = COMPLEMENT;
  expr_stack[expr_depth] = expr;
  expr_depth++;
/* 
  print_expression_type(expr_stack[expr_depth]->type);
  ansi_fflush(stderr);
*/
}
  


void
store_convex_pred_expression() {
  expr_stack[expr_depth]->pred = convex_pred;
}

/* this is for the staetments occurring in a recursion/iteration expr */
/* statements should already be stored as linked list in correct position */
/* only need to pop off and put in the expression */
void
store_stmts_in_expr() {
  expr_stack[expr_depth]->stmt = stmt_stack[stmt_stack_nr];
  stmt_stack[stmt_stack_nr] = NULL;
  stmt_stack_nr--;
  stmt_depth--;
/*
fprintf(stderr,"after storing in expr, current stmt stack nr = %d",stmt_stack_nr);
print_expression_type(expr_stack[expr_depth]->type);
ansi_fflush(stderr);
ansi_fflush(stdout);  
*/
}

/* ==================== basic Bool epxressions ======= */

/* exprtype = 0 for false, 1 for true, -1 for inequality */
void
init_Boolexpression(enum Boolexpr_type bet) {
  Boolexpr_depth++;
  Boolexpr_stack[Boolexpr_depth] = get_new_Boolexpression();
  Boolexpr_stack[Boolexpr_depth]->type = bet;
}

/* assume have already initialized a new Boolexpression on the stack */
void
store_ineq_Boolexpression() {
  Boolexpr_stack[Boolexpr_depth]->left_expr =  expr_stack[expr_depth-1];
  Boolexpr_stack[Boolexpr_depth]->right_expr = expr_stack[expr_depth];
  Boolexpr_stack[Boolexpr_depth]->type = BE_INEQ;
  expr_depth -= 2;
}
 

void
store_or_Boolexpression(){
  struct Boolexpression *Boolexpr = get_new_Boolexpression();
  Boolexpr->Boolexpr1 =  Boolexpr_stack[ Boolexpr_depth-1];
  Boolexpr->Boolexpr2 =  Boolexpr_stack[ Boolexpr_depth];
  Boolexpr->type = BE_OR;
  Boolexpr_depth--;
  Boolexpr_stack[Boolexpr_depth] = Boolexpr;
}

void
store_and_Boolexpression(){
  struct Boolexpression *Boolexpr = get_new_Boolexpression();
  Boolexpr->Boolexpr1 =  Boolexpr_stack[ Boolexpr_depth-1];
  Boolexpr->Boolexpr2 =  Boolexpr_stack[ Boolexpr_depth];
  Boolexpr->type = BE_AND; 
  Boolexpr_depth--;
  Boolexpr_stack[Boolexpr_depth] = Boolexpr;
}


void
store_not_Boolexpression() {
  struct Boolexpression *Boolexpr = get_new_Boolexpression();
  Boolexpr->Boolexpr1 =  Boolexpr_stack[ Boolexpr_depth];
  Boolexpr->type = BE_NOT; 
  Boolexpr_stack[Boolexpr_depth] = Boolexpr;
}


void
store_expr_in_empty_Boolexpression() {
  struct Boolexpression *Boolexpr = get_new_Boolexpression();
  Boolexpr->left_expr =  expr_stack[ expr_depth];
  expr_stack[ expr_depth] = NULL;
  expr_depth--;
  Boolexpr->type = BE_EMPTY; 
  /* Boolexpr_depth++; added 8/14/00 to fix bug */
  Boolexpr_depth++;
  Boolexpr_stack[Boolexpr_depth] = Boolexpr;
}




/* routines mirrors store_stmt_in_expr() */
/* this is for the staetments occurring in an IF ... stmt */
/* statements should already be stored as linked list in correct position */
/* only need to pop off and put in the expression */
void
store_stmt_in_stmt(enum stmt_type typ) {
  /* the if stmt is one below the current TOS */
  switch (typ) {
  case IF_THEN_STMT:
    stmt_stack[stmt_stack_nr-1]->thenstmt = stmt_stack[stmt_stack_nr];
    break;
  case IF_THEN_ELSE_STMT: /* really means the else part */
    stmt_stack[stmt_stack_nr-1]->elsestmt = stmt_stack[stmt_stack_nr];
    break;
  case WHILE_STMT:
    stmt_stack[stmt_stack_nr-1]->loopstmt = stmt_stack[stmt_stack_nr];
    break;
  default:
    fprintf(stderr,"Illegal call of store_stmt_in_stmt\n");ansi_fflush(stderr);
  }
  stmt_stack[stmt_stack_nr-1]->type = typ;
  stmt_stack[stmt_stack_nr] = NULL;
  stmt_stack_nr--;
  stmt_depth--;
}



/* takes Boolexpr off top of stack and puts in statement at TOS */
void
store_Boolexpr_in_stmt() {
  stmt_stack[stmt_stack_nr]->Boolexpr = Boolexpr_stack[Boolexpr_depth];
  Boolexpr_stack[Boolexpr_depth] = NULL;
  Boolexpr_depth--;
}

void 
store_ineq_type_in_Boolexpression(int relop_type) {
  Boolexpr_stack[Boolexpr_depth]->optype = relop_type;
}

